"""
The pygad.utils.crossover module has all the built-in crossover operators.
"""

import numpy
import random

class Crossover:

    def __init__():
        pass

    def single_point_crossover(self, parents, offspring_size):

        """
        Applies single-point crossover between pairs of parents.
        This function selects a random point at which crossover occurs between the parents, generating offspring.

        Parameters:
            parents (array-like): The parents to mate for producing the offspring.
            offspring_size (int): The number of offspring to produce.

        Returns:
            array-like: An array containing the produced offspring.
        """

        if self.gene_type_single == True:
            offspring = numpy.empty(offspring_size, dtype=self.gene_type[0])
        else:
            offspring = numpy.empty(offspring_size, dtype=object)

        # Randomly generate all the K points at which crossover takes place between each two parents. The point does not have to be always at the center of the solutions.
        # This saves time by calling the numpy.random.randint() function only once.
        crossover_points = numpy.random.randint(low=0, 
                                                high=parents.shape[1], 
                                                size=offspring_size[0])

        for k in range(offspring_size[0]):
            # Check if the crossover_probability parameter is used.
            if not (self.crossover_probability is None):
                probs = numpy.random.random(size=parents.shape[0])
                indices = list(set(numpy.where(probs <= self.crossover_probability)[0]))

                # If no parent satisfied the probability, no crossover is applied and a parent is selected as is.
                if len(indices) == 0:
                    offspring[k, :] = parents[k % parents.shape[0], :]
                    continue
                elif len(indices) == 1:
                    parent1_idx = indices[0]
                    parent2_idx = parent1_idx
                else:
                    indices = random.sample(indices, 2)
                    parent1_idx = indices[0]
                    parent2_idx = indices[1]
            else:
                # Index of the first parent to mate.
                parent1_idx = k % parents.shape[0]
                # Index of the second parent to mate.
                parent2_idx = (k+1) % parents.shape[0]

            # The new offspring has its first half of its genes from the first parent.
            offspring[k, 0:crossover_points[k]] = parents[parent1_idx, 0:crossover_points[k]]
            # The new offspring has its second half of its genes from the second parent.
            offspring[k, crossover_points[k]:] = parents[parent2_idx, crossover_points[k]:]

            if self.allow_duplicate_genes == False:
                if self.gene_space is None:
                    offspring[k], _, _ = self.solve_duplicate_genes_randomly(solution=offspring[k],
                                                                             min_val=self.random_mutation_min_val,
                                                                             max_val=self.random_mutation_max_val,
                                                                             mutation_by_replacement=self.mutation_by_replacement,
                                                                             gene_type=self.gene_type,
                                                                             sample_size=self.sample_size)
                else:
                    offspring[k], _, _ = self.solve_duplicate_genes_by_space(solution=offspring[k],
                                                                             gene_type=self.gene_type,
                                                                             sample_size=self.sample_size,
                                                                             mutation_by_replacement=self.mutation_by_replacement,
                                                                             build_initial_pop=False)
        
        return offspring

    def two_points_crossover(self, parents, offspring_size):

        """
        Applies the 2 points crossover. It selects the 2 points randomly at which crossover takes place between the pairs of parents.
        It accepts 2 parameters:
            -parents: The parents to mate for producing the offspring.
            -offspring_size: The size of the offspring to produce.
        It returns an array the produced offspring.
        """

        if self.gene_type_single == True:
            offspring = numpy.empty(offspring_size, dtype=self.gene_type[0])
        else:
            offspring = numpy.empty(offspring_size, dtype=object)

        # Randomly generate all the first K points at which crossover takes place between each two parents. 
        # This saves time by calling the numpy.random.randint() function only once.
        if (parents.shape[1] == 1): # If the chromosome has only a single gene. In this case, this gene is copied from the second parent.
            crossover_points_1 = numpy.zeros(offspring_size[0])
        else:
            crossover_points_1 = numpy.random.randint(low=0, 
                                                      high=numpy.ceil(parents.shape[1]/2 + 1), 
                                                      size=offspring_size[0])

        # The second point must always be greater than the first point.
        crossover_points_2 = crossover_points_1 + int(parents.shape[1]/2) 

        for k in range(offspring_size[0]):

            if not (self.crossover_probability is None):
                probs = numpy.random.random(size=parents.shape[0])
                indices = list(set(numpy.where(probs <= self.crossover_probability)[0]))

                # If no parent satisfied the probability, no crossover is applied and a parent is selected.
                if len(indices) == 0:
                    offspring[k, :] = parents[k % parents.shape[0], :]
                    continue
                elif len(indices) == 1:
                    parent1_idx = indices[0]
                    parent2_idx = parent1_idx
                else:
                    indices = random.sample(indices, 2)
                    parent1_idx = indices[0]
                    parent2_idx = indices[1]
            else:
                # Index of the first parent to mate.
                parent1_idx = k % parents.shape[0]
                # Index of the second parent to mate.
                parent2_idx = (k+1) % parents.shape[0]

            # The genes from the beginning of the chromosome up to the first point are copied from the first parent.
            offspring[k, 0:crossover_points_1[k]] = parents[parent1_idx, 0:crossover_points_1[k]]
            # The genes from the second point up to the end of the chromosome are copied from the first parent.
            offspring[k, crossover_points_2[k]:] = parents[parent1_idx, crossover_points_2[k]:]
            # The genes between the 2 points are copied from the second parent.
            offspring[k, crossover_points_1[k]:crossover_points_2[k]] = parents[parent2_idx, crossover_points_1[k]:crossover_points_2[k]]

            if self.allow_duplicate_genes == False:
                if self.gene_space is None:
                    offspring[k], _, _ = self.solve_duplicate_genes_randomly(solution=offspring[k],
                                                                             min_val=self.random_mutation_min_val,
                                                                             max_val=self.random_mutation_max_val,
                                                                             mutation_by_replacement=self.mutation_by_replacement,
                                                                             gene_type=self.gene_type,
                                                                             sample_size=self.sample_size)
                else:
                    offspring[k], _, _ = self.solve_duplicate_genes_by_space(solution=offspring[k],
                                                                             gene_type=self.gene_type,
                                                                             sample_size=self.sample_size,
                                                                             mutation_by_replacement=self.mutation_by_replacement,
                                                                             build_initial_pop=False)
        return offspring

    def uniform_crossover(self, parents, offspring_size):

        """
        Applies the uniform crossover. For each gene, a parent out of the 2 mating parents is selected randomly and the gene is copied from it.
        It accepts 2 parameters:
            -parents: The parents to mate for producing the offspring.
            -offspring_size: The size of the offspring to produce.
        It returns an array the produced offspring.
        """

        if self.gene_type_single == True:
            offspring = numpy.empty(offspring_size, dtype=self.gene_type[0])
        else:
            offspring = numpy.empty(offspring_size, dtype=object)

        # Randomly generate all the genes sources at which crossover takes place between each two parents. 
        # This saves time by calling the numpy.random.randint() function only once.
        # There is a list of 0 and 1 for each offspring.
        # [0, 1, 0, 0, 1, 1]: If the value is 0, then take the gene from the first parent. If 1, take it from the second parent.
        genes_sources = numpy.random.randint(low=0, 
                                             high=2, 
                                             size=offspring_size)

        for k in range(offspring_size[0]):
            if not (self.crossover_probability is None):
                probs = numpy.random.random(size=parents.shape[0])
                indices = list(set(numpy.where(probs <= self.crossover_probability)[0]))

                # If no parent satisfied the probability, no crossover is applied and a parent is selected.
                if len(indices) == 0:
                    offspring[k, :] = parents[k % parents.shape[0], :]
                    continue
                elif len(indices) == 1:
                    parent1_idx = indices[0]
                    parent2_idx = parent1_idx
                else:
                    indices = random.sample(indices, 2)
                    parent1_idx = indices[0]
                    parent2_idx = indices[1]
            else:
                # Index of the first parent to mate.
                parent1_idx = k % parents.shape[0]
                # Index of the second parent to mate.
                parent2_idx = (k+1) % parents.shape[0]

            for gene_idx in range(offspring_size[1]):
                if (genes_sources[k, gene_idx] == 0):
                    # The gene will be copied from the first parent if the current gene index is 0.
                    offspring[k, gene_idx] = parents[parent1_idx, gene_idx]
                elif (genes_sources[k, gene_idx] == 1):
                    # The gene will be copied from the second parent if the current gene index is 1.
                    offspring[k, gene_idx] = parents[parent2_idx, gene_idx]

            if self.allow_duplicate_genes == False:
                if self.gene_space is None:
                    offspring[k], _, _ = self.solve_duplicate_genes_randomly(solution=offspring[k],
                                                                             min_val=self.random_mutation_min_val,
                                                                             max_val=self.random_mutation_max_val,
                                                                             mutation_by_replacement=self.mutation_by_replacement,
                                                                             gene_type=self.gene_type,
                                                                             sample_size=self.sample_size)
                else:
                    offspring[k], _, _ = self.solve_duplicate_genes_by_space(solution=offspring[k],
                                                                             gene_type=self.gene_type,
                                                                             sample_size=self.sample_size,
                                                                             mutation_by_replacement=self.mutation_by_replacement,
                                                                             build_initial_pop=False)

        return offspring

    def scattered_crossover(self, parents, offspring_size):

        """
        Applies the scattered crossover. It randomly selects the gene from one of the 2 parents. 
        It accepts 2 parameters:
            -parents: The parents to mate for producing the offspring.
            -offspring_size: The size of the offspring to produce.
        It returns an array the produced offspring.
        """

        if self.gene_type_single == True:
            offspring = numpy.empty(offspring_size, dtype=self.gene_type[0])
        else:
            offspring = numpy.empty(offspring_size, dtype=object)

        # Randomly generate all the genes sources at which crossover takes place between each two parents. 
        # This saves time by calling the numpy.random.randint() function only once.
        # There is a list of 0 and 1 for each offspring.
        # [0, 1, 0, 0, 1, 1]: If the value is 0, then take the gene from the first parent. If 1, take it from the second parent.
        genes_sources = numpy.random.randint(low=0, 
                                             high=2, 
                                             size=offspring_size)

        for k in range(offspring_size[0]):
            if not (self.crossover_probability is None):
                probs = numpy.random.random(size=parents.shape[0])
                indices = list(set(numpy.where(probs <= self.crossover_probability)[0]))

                # If no parent satisfied the probability, no crossover is applied and a parent is selected.
                if len(indices) == 0:
                    offspring[k, :] = parents[k % parents.shape[0], :]
                    continue
                elif len(indices) == 1:
                    parent1_idx = indices[0]
                    parent2_idx = parent1_idx
                else:
                    indices = random.sample(indices, 2)
                    parent1_idx = indices[0]
                    parent2_idx = indices[1]
            else:
                # Index of the first parent to mate.
                parent1_idx = k % parents.shape[0]
                # Index of the second parent to mate.
                parent2_idx = (k+1) % parents.shape[0]

            offspring[k, :] = numpy.where(genes_sources[k] == 0, 
                                          parents[parent1_idx, :], 
                                          parents[parent2_idx, :])

            if self.allow_duplicate_genes == False:
                if self.gene_space is None:
                    offspring[k], _, _ = self.solve_duplicate_genes_randomly(solution=offspring[k],
                                                                             min_val=self.random_mutation_min_val,
                                                                             max_val=self.random_mutation_max_val,
                                                                             mutation_by_replacement=self.mutation_by_replacement,
                                                                             gene_type=self.gene_type,
                                                                             sample_size=self.sample_size)
                else:
                    offspring[k], _, _ = self.solve_duplicate_genes_by_space(solution=offspring[k],
                                                                             gene_type=self.gene_type,
                                                                             sample_size=self.sample_size,
                                                                             mutation_by_replacement=self.mutation_by_replacement,
                                                                             build_initial_pop=False)
        return offspring



import pygad
import random
import numpy

num_generations = 1

initial_population = [[0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                      [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                      [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                      [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                      [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                      [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                      [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                      [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                      [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                      [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]]

def population_gene_constraint(gene_space=None,
                               gene_type=float,
                               num_genes=10,
                               mutation_by_replacement=False,
                               random_mutation_min_val=-1,
                               random_mutation_max_val=1,
                               init_range_low=-4,
                               init_range_high=4,
                               random_seed=123,
                               crossover_type='single_point',
                               initial_population=None,
                               parent_selection_type='sss',
                               multi_objective=False,
                               gene_constraint=None,
                               allow_duplicate_genes=True):

    def fitness_func_no_batch_single(ga, solution, idx):
        return random.random()

    def fitness_func_no_batch_multi(ga, solution, idx):
        return [random.random(), random.random()]

    if multi_objective == True:
        fitness_func = fitness_func_no_batch_multi
    else:
        fitness_func = fitness_func_no_batch_single

    ga_instance = pygad.GA(num_generations=num_generations,
                           num_parents_mating=5,
                           fitness_func=fitness_func,
                           sol_per_pop=10,
                           num_genes=num_genes,
                           gene_space=gene_space,
                           gene_type=gene_type,
                           initial_population=initial_population,
                           parent_selection_type=parent_selection_type,
                           init_range_low=init_range_low,
                           init_range_high=init_range_high,
                           random_mutation_min_val=random_mutation_min_val,
                           random_mutation_max_val=random_mutation_max_val,
                           allow_duplicate_genes=allow_duplicate_genes,
                           mutation_by_replacement=mutation_by_replacement,
                           random_seed=random_seed,
                           crossover_type=crossover_type,
                           gene_constraint=gene_constraint,
                           save_solutions=True,
                           suppress_warnings=False)

    ga_instance.run()

    return ga_instance

#### Single-Objective
def test_initial_population_int_by_replacement():
    gene_constraint=[lambda x,v: [val for val in v if val>=8],
                     lambda x,v: [val for val in v if val>=8],
                     lambda x,v: [val for val in v if 5>=val>=1],
                     lambda x,v: [val for val in v if 5>val>3],
                     lambda x,v: [val for val in v if val<2]]
    ga_instance = population_gene_constraint(gene_constraint=gene_constraint,
                                             init_range_low=0,
                                             init_range_high=10,
                                             random_mutation_min_val=0,
                                             random_mutation_max_val=10,
                                             num_genes=5,
                                             gene_type=int,
                                             mutation_by_replacement=True)
    initial_population = ga_instance.initial_population

    assert numpy.all(initial_population[:, 0] >= 8), "Not all values in column 0 are >= 8"
    assert numpy.all(initial_population[:, 1] >= 8), "Not all values in column 1 are >= 8"
    assert numpy.all(initial_population[:, 2] >= 1), "Not all values in column 2 are >= 1"
    assert numpy.all((initial_population[:, 2] >= 1) & (initial_population[:, 2] <= 5)), "Not all values in column 2 between 1 and 5 (inclusive)"
    assert numpy.all(initial_population[:, 3] == 4), "Not all values in column 3 between 3 and 5 (exclusive)"
    assert numpy.all(initial_population[:, 4] < 2), "Not all values in column 4 < 2"
    print("All constraints are met")

def test_initial_population_int_by_replacement_no_duplicates():
    gene_constraint=[lambda x,v: [val for val in v if val>=5],
                     lambda x,v: [val for val in v if val>=5],
                     lambda x,v: [val for val in v if val>=5],
                     lambda x,v: [val for val in v if val>=5],
                     lambda x,v: [val for val in v if val>=5]]
    ga_instance = population_gene_constraint(gene_constraint=gene_constraint,
                                             init_range_low=1,
                                             init_range_high=10,
                                             random_mutation_min_val=1,
                                             random_mutation_max_val=10,
                                             gene_type=int,
                                             num_genes=5,
                                             mutation_by_replacement=True,
                                             allow_duplicate_genes=False)

    num_duplicates = 0
    for idx, solution in enumerate(ga_instance.solutions):
        num = len(solution) - len(set(solution))
        if num != 0:
            print(solution, idx)
        num_duplicates += num

    assert num_duplicates == 0

    initial_population = ga_instance.initial_population
    # print(initial_population)

    assert numpy.all(initial_population[:, 0] >= 5), "Not all values in column 0 >= 5"
    assert numpy.all(initial_population[:, 1] >= 5), "Not all values in column 1 >= 5"
    assert numpy.all(initial_population[:, 2] >= 5), "Not all values in column 2 >= 5"
    assert numpy.all(initial_population[:, 3] >= 5), "Not all values in column 3 >= 5"
    assert numpy.all(initial_population[:, 4] >= 5), "Not all values in column 4 >= 5"
    print("All constraints are met")

def test_initial_population_int_by_replacement_no_duplicates2():
    gene_constraint=[lambda x,v: [val for val in v if val>=98],
                     lambda x,v: [val for val in v if val>=98],
                     lambda x,v: [val for val in v if 20<val<40],
                     lambda x,v: [val for val in v if val<40],
                     lambda x,v: [val for val in v if val<50],
                     lambda x,v: [val for val in v if val<100]]
    ga_instance = population_gene_constraint(gene_constraint=gene_constraint,
                                             random_mutation_min_val=1,
                                             random_mutation_max_val=100,
                                             init_range_low=1,
                                             init_range_high=100,
                                             gene_type=int,
                                             num_genes=6,
                                             crossover_type=None,
                                             mutation_by_replacement=True,
                                             allow_duplicate_genes=False)

    num_duplicates = 0
    for idx, solution in enumerate(ga_instance.solutions):
        num = len(solution) - len(set(solution))
        if num != 0:
            print(solution, idx)
        num_duplicates += num

    assert num_duplicates == 0

    initial_population = ga_instance.initial_population
    # print(initial_population)

    assert numpy.all(initial_population[:, 0] >= 98), "Not all values in column 0 are >= 98"
    assert numpy.all(initial_population[:, 1] >= 98), "Not all values in column 1 are >= 98"
    assert numpy.all((initial_population[:, 2] > 20) & (initial_population[:, 2] < 40)), "Not all values in column 2 between 20 and 40 (exclusive)"
    assert numpy.all(initial_population[:, 3] < 40), "Not all values in column 3 < 40"
    assert numpy.all(initial_population[:, 4] < 50), "Not all values in column 4 < 50"
    assert numpy.all(initial_population[:, 5] < 100), "Not all values in column 4 < 100"
    print("All constraints are met")

def test_initial_population_float_by_replacement_no_duplicates():
    gene_constraint=[lambda x,v: [val for val in v if val>=5],
                     lambda x,v: [val for val in v if val>=5],
                     lambda x,v: [val for val in v if val>=5],
                     lambda x,v: [val for val in v if val>=5],
                     lambda x,v: [val for val in v if val>=5]]
    ga_instance = population_gene_constraint(gene_constraint=gene_constraint,
                                             init_range_low=1,
                                             init_range_high=10,
                                             gene_type=[float, 1],
                                             num_genes=5,
                                             crossover_type=None,
                                             mutation_by_replacement=False,
                                             allow_duplicate_genes=False)

    num_duplicates = 0
    for idx, solution in enumerate(ga_instance.solutions):
        num = len(solution) - len(set(solution))
        if num != 0:
            print(solution, idx)
        num_duplicates += num

    assert num_duplicates == 0

    initial_population = ga_instance.initial_population
    # print(initial_population)

    assert numpy.all(initial_population[:, 0] >= 5), "Not all values in column 0 >= 5"
    assert numpy.all(initial_population[:, 1] >= 5), "Not all values in column 1 >= 5"
    assert numpy.all(initial_population[:, 2] >= 5), "Not all values in column 2 >= 5"
    assert numpy.all(initial_population[:, 3] >= 5), "Not all values in column 3 >= 5"
    assert numpy.all(initial_population[:, 4] >= 5), "Not all values in column 4 >= 5"
    print("All constraints are met")

def test_initial_population_float_by_replacement_no_duplicates2():
    gene_constraint=[lambda x,v: [val for val in v if val>=1],
                     lambda x,v: [val for val in v if val>=1],
                     lambda x,v: [val for val in v if val>=1],
                     lambda x,v: [val for val in v if val>=1],
                     lambda x,v: [val for val in v if val>=1]]
    ga_instance = population_gene_constraint(gene_constraint=gene_constraint,
                                             init_range_low=1,
                                             init_range_high=2,
                                             gene_type=[float, 1],
                                             num_genes=5,
                                             crossover_type=None,
                                             mutation_by_replacement=False,
                                             allow_duplicate_genes=False)

    num_duplicates = 0
    for idx, solution in enumerate(ga_instance.solutions):
        num = len(solution) - len(set(solution))
        if num != 0:
            print(solution, idx)
        num_duplicates += num

    assert num_duplicates == 0

    initial_population = ga_instance.initial_population
    # print(initial_population)

    assert numpy.all(initial_population[:, 0] >= 1), "Not all values in column 0 >= 1"
    assert numpy.all(initial_population[:, 1] >= 1), "Not all values in column 1 >= 1"
    assert numpy.all(initial_population[:, 2] >= 1), "Not all values in column 2 >= 1"
    assert numpy.all(initial_population[:, 3] >= 1), "Not all values in column 3 >= 1"
    assert numpy.all(initial_population[:, 4] >= 1), "Not all values in column 4 >= 1"
    print("All constraints are met")

def test_initial_population_float_by_replacement_no_duplicates_None_constraints():
    gene_constraint=[lambda x,v: [val for val in v if val>=1],
                     None,
                     lambda x,v: [val for val in v if val>=1],
                     None,
                     lambda x,v: [val for val in v if val>=1]]
    ga_instance = population_gene_constraint(gene_constraint=gene_constraint,
                                             init_range_low=1,
                                             init_range_high=2,
                                             gene_type=[float, 1],
                                             num_genes=5,
                                             crossover_type=None,
                                             mutation_by_replacement=False,
                                             allow_duplicate_genes=False)

    num_duplicates = 0
    for idx, solution in enumerate(ga_instance.solutions):
        num = len(solution) - len(set(solution))
        if num != 0:
            print(solution, idx)
        num_duplicates += num

    assert num_duplicates == 0

    initial_population = ga_instance.initial_population
    # print(initial_population)

    assert numpy.all(initial_population[:, 0] >= 1), "Not all values in column 0 >= 1"
    assert numpy.all(initial_population[:, 1] >= 1), "Not all values in column 1 >= 1"
    assert numpy.all(initial_population[:, 2] >= 1), "Not all values in column 2 >= 1"
    assert numpy.all(initial_population[:, 3] >= 1), "Not all values in column 3 >= 1"
    assert numpy.all(initial_population[:, 4] >= 1), "Not all values in column 4 >= 1"
    print("All constraints are met")

if __name__ == "__main__":
    #### Single-objective
    print()
    test_initial_population_int_by_replacement()
    print()
    test_initial_population_int_by_replacement_no_duplicates()
    print()
    test_initial_population_int_by_replacement_no_duplicates2()
    print()
    test_initial_population_float_by_replacement_no_duplicates()
    print()
    test_initial_population_float_by_replacement_no_duplicates2()
    print()
    test_initial_population_float_by_replacement_no_duplicates_None_constraints()
    print()



import pygad
import random
import numpy

actual_num_fitness_calls_default_keep = 0
actual_num_fitness_calls_no_keep = 0
actual_num_fitness_calls_keep_elitism = 0
actual_num_fitness_calls_keep_parents = 0

num_generations = 100
sol_per_pop = 10
num_parents_mating = 5

# TODO: Calculate the number when fitness_batch_size is used.

def number_calls_fitness_function(keep_elitism=1, 
                                  keep_parents=-1,
                                  mutation_type="random",
                                  mutation_percent_genes="default",
                                  parent_selection_type='sss',
                                  multi_objective=False,
                                  fitness_batch_size=None):

    actual_num_fitness_calls = 0
    def fitness_func_no_batch_single(ga, solution, idx):
        nonlocal actual_num_fitness_calls
        actual_num_fitness_calls = actual_num_fitness_calls + 1
        return random.random()

    def fitness_func_no_batch_multi(ga_instance, solution, solution_idx):
        nonlocal actual_num_fitness_calls
        actual_num_fitness_calls = actual_num_fitness_calls + 1
        return [random.random(), random.random()]

    def fitness_func_batch_single(ga_instance, solution, solution_idx):
        nonlocal actual_num_fitness_calls
        actual_num_fitness_calls = actual_num_fitness_calls + 1
        f = []
        for sol in solution:
            f.append(random.random())
        return f

    def fitness_func_batch_multi(ga_instance, solution, solution_idx):
        nonlocal actual_num_fitness_calls
        actual_num_fitness_calls = actual_num_fitness_calls + 1
        f = []
        for sol in solution:
            f.append([random.random(), random.random()])
        return f

    if fitness_batch_size is None or (type(fitness_batch_size) in pygad.GA.supported_int_types and fitness_batch_size == 1):
        if multi_objective == True:
            fitness_func = fitness_func_no_batch_multi
        else:
            fitness_func = fitness_func_no_batch_single
    elif (type(fitness_batch_size) in pygad.GA.supported_int_types and fitness_batch_size > 1):
        if multi_objective == True:
            fitness_func = fitness_func_batch_multi
        else:
            fitness_func = fitness_func_batch_single

    ga_optimizer = pygad.GA(num_generations=num_generations,
                            sol_per_pop=sol_per_pop,
                            num_genes=6,
                            num_parents_mating=num_parents_mating,
                            fitness_func=fitness_func,
                            mutation_type=mutation_type,
                            parent_selection_type=parent_selection_type,
                            mutation_percent_genes=mutation_percent_genes,
                            keep_elitism=keep_elitism,
                            keep_parents=keep_parents,
                            suppress_warnings=True,
                            fitness_batch_size=fitness_batch_size)

    ga_optimizer.run()

    if fitness_batch_size is None:
        if keep_elitism == 0:
            if keep_parents == 0:
                # 10 (for initial population) + 100*10 (for other generations) = 1010
                expected_num_fitness_calls = sol_per_pop + num_generations * sol_per_pop
                if mutation_type == "adaptive":
                    expected_num_fitness_calls += num_generations * sol_per_pop
            elif keep_parents == -1:
                # 10 (for initial population) + 100*num_parents_mating (for other generations)
                expected_num_fitness_calls = sol_per_pop + num_generations * (sol_per_pop - num_parents_mating)
                if mutation_type == "adaptive":
                    expected_num_fitness_calls += num_generations * (sol_per_pop - num_parents_mating)
            else:
                # 10 (for initial population) + 100*keep_parents (for other generations)
                expected_num_fitness_calls = sol_per_pop + num_generations * (sol_per_pop - keep_parents)
                if mutation_type == "adaptive":
                    expected_num_fitness_calls += num_generations * (sol_per_pop - keep_parents)
        else:
            # 10 (for initial population) + 100*keep_elitism (for other generations)
            expected_num_fitness_calls = sol_per_pop + num_generations * (sol_per_pop - keep_elitism)
            if mutation_type == "adaptive":
                expected_num_fitness_calls += num_generations * (sol_per_pop - keep_elitism)
    else:
        if keep_elitism == 0:
            if keep_parents == 0:
                # 10 (for initial population) + 100*10 (for other generations) = 1010
                expected_num_fitness_calls = int(numpy.ceil(sol_per_pop/fitness_batch_size)) + num_generations * int(numpy.ceil(sol_per_pop/fitness_batch_size))
                if mutation_type == "adaptive":
                    expected_num_fitness_calls += num_generations * int(numpy.ceil(sol_per_pop/fitness_batch_size))
            elif keep_parents == -1:
                # 10 (for initial population) + 100*num_parents_mating (for other generations)
                expected_num_fitness_calls = int(numpy.ceil(sol_per_pop/fitness_batch_size)) + num_generations * int(numpy.ceil((sol_per_pop - num_parents_mating)/fitness_batch_size))
                if mutation_type == "adaptive":
                    expected_num_fitness_calls += num_generations * int(numpy.ceil((sol_per_pop - num_parents_mating)/fitness_batch_size))
            else:
                # 10 (for initial population) + 100*keep_parents (for other generations)
                expected_num_fitness_calls = int(numpy.ceil(sol_per_pop/fitness_batch_size)) + num_generations * int(numpy.ceil((sol_per_pop - keep_parents)/fitness_batch_size))
                if mutation_type == "adaptive":
                    expected_num_fitness_calls += num_generations * int(numpy.ceil((sol_per_pop - keep_parents)/fitness_batch_size))
        else:
            # 10 (for initial population) + 100*keep_elitism (for other generations)
            expected_num_fitness_calls = int(numpy.ceil(sol_per_pop/fitness_batch_size)) + num_generations * int(numpy.ceil((sol_per_pop - keep_elitism)/fitness_batch_size))
            if mutation_type == "adaptive":
                expected_num_fitness_calls += num_generations * int(numpy.ceil((sol_per_pop - keep_elitism)/fitness_batch_size))

    print(f"Expected number of fitness function calls is {expected_num_fitness_calls}.")
    print(f"Actual number of fitness function calls is {actual_num_fitness_calls}.")
    return actual_num_fitness_calls, expected_num_fitness_calls

def test_number_calls_fitness_function_default_keep():
    actual, expected = number_calls_fitness_function()
    assert actual == expected

def test_number_calls_fitness_function_no_keep():
    actual, expected = number_calls_fitness_function(keep_elitism=0, 
                                                     keep_parents=0)
    assert actual == expected

def test_number_calls_fitness_function_keep_elitism():
    actual, expected = number_calls_fitness_function(keep_elitism=3, 
                                                     keep_parents=0)
    assert actual == expected

def test_number_calls_fitness_function_keep_parents():
    actual, expected = number_calls_fitness_function(keep_elitism=0, 
                                                     keep_parents=4)
    assert actual == expected

def test_number_calls_fitness_function_both_keep():
    actual, expected = number_calls_fitness_function(keep_elitism=3, 
                                                     keep_parents=4)
    assert actual == expected

def test_number_calls_fitness_function_no_keep_adaptive_mutation():
    actual, expected = number_calls_fitness_function(keep_elitism=0, 
                                                     keep_parents=0,
                                                     mutation_type="adaptive",
                                                     mutation_percent_genes=[10, 5])
    assert actual == expected

def test_number_calls_fitness_function_default_adaptive_mutation():
    actual, expected = number_calls_fitness_function(mutation_type="adaptive",
                                                     mutation_percent_genes=[10, 5])
    assert actual == expected

def test_number_calls_fitness_function_both_keep_adaptive_mutation():
    actual, expected = number_calls_fitness_function(keep_elitism=3, 
                                                     keep_parents=4,
                                                     mutation_type="adaptive",
                                                     mutation_percent_genes=[10, 5])
    assert actual == expected

#### Multi Objective
def test_number_calls_fitness_function_no_keep_multi_objective():
    actual, expected = number_calls_fitness_function(keep_elitism=0, 
                                                     keep_parents=0,
                                                     multi_objective=True)
    assert actual == expected

def test_number_calls_fitness_function_keep_elitism_multi_objective():
    actual, expected = number_calls_fitness_function(keep_elitism=3, 
                                                     keep_parents=0,
                                                     multi_objective=True)
    assert actual == expected

def test_number_calls_fitness_function_keep_parents_multi_objective():
    actual, expected = number_calls_fitness_function(keep_elitism=0, 
                                                     keep_parents=4,
                                                     multi_objective=True)
    assert actual == expected

def test_number_calls_fitness_function_both_keep_multi_objective():
    actual, expected = number_calls_fitness_function(keep_elitism=3, 
                                                     keep_parents=4,
                                                     multi_objective=True)
    assert actual == expected

def test_number_calls_fitness_function_no_keep_adaptive_mutation_multi_objective():
    actual, expected = number_calls_fitness_function(keep_elitism=0, 
                                                     keep_parents=0,
                                                     mutation_type="adaptive",
                                                     mutation_percent_genes=[10, 5],
                                                     multi_objective=True)
    assert actual == expected

def test_number_calls_fitness_function_default_adaptive_mutation_multi_objective():
    actual, expected = number_calls_fitness_function(mutation_type="adaptive",
                                                     mutation_percent_genes=[10, 5],
                                                     multi_objective=True)
    assert actual == expected

def test_number_calls_fitness_function_both_keep_adaptive_mutation_multi_objective():
    actual, expected = number_calls_fitness_function(keep_elitism=3, 
                                                     keep_parents=4,
                                                     mutation_type="adaptive",
                                                     mutation_percent_genes=[10, 5],
                                                     multi_objective=True)
    assert actual == expected

#### Multi Objective NSGA-II Parent Selection
def test_number_calls_fitness_function_no_keep_multi_objective_nsga2():
    actual, expected = number_calls_fitness_function(keep_elitism=0, 
                                                     keep_parents=0,
                                                     multi_objective=True)
    assert actual == expected

def test_number_calls_fitness_function_keep_elitism_multi_objective_nsga2():
    actual, expected = number_calls_fitness_function(keep_elitism=3, 
                                                     keep_parents=0,
                                                     multi_objective=True)
    assert actual == expected

def test_number_calls_fitness_function_keep_parents_multi_objective_nsga2():
    actual, expected = number_calls_fitness_function(keep_elitism=0, 
                                                     keep_parents=4,
                                                     multi_objective=True)
    assert actual == expected

def test_number_calls_fitness_function_both_keep_multi_objective_nsga2():
    actual, expected = number_calls_fitness_function(keep_elitism=3, 
                                                     keep_parents=4,
                                                     multi_objective=True)
    assert actual == expected

def test_number_calls_fitness_function_no_keep_adaptive_mutation_multi_objective_nsga2():
    actual, expected = number_calls_fitness_function(keep_elitism=0, 
                                                     keep_parents=0,
                                                     mutation_type="adaptive",
                                                     mutation_percent_genes=[10, 5],
                                                     multi_objective=True)
    assert actual == expected

def test_number_calls_fitness_function_default_adaptive_mutation_multi_objective_nsga2():
    actual, expected = number_calls_fitness_function(mutation_type="adaptive",
                                                     mutation_percent_genes=[10, 5],
                                                     multi_objective=True)
    assert actual == expected

def test_number_calls_fitness_function_both_keep_adaptive_mutation_multi_objective_nsga2():
    actual, expected = number_calls_fitness_function(keep_elitism=3, 
                                                     keep_parents=4,
                                                     mutation_type="adaptive",
                                                     mutation_percent_genes=[10, 5],
                                                     multi_objective=True)
    assert actual == expected


######## Batch Fitness Calculation
#### Single Objective
def test_number_calls_fitness_function_no_keep_batch_1():
    actual, expected = number_calls_fitness_function(fitness_batch_size=1)
    assert actual == expected

def test_number_calls_fitness_function_no_keep_batch_4():
    actual, expected = number_calls_fitness_function(fitness_batch_size=4)
    assert actual == expected

def test_number_calls_fitness_function_no_keep_batch_9():
    actual, expected = number_calls_fitness_function(fitness_batch_size=9)
    assert actual == expected

def test_number_calls_fitness_function_no_keep_batch_10():
    actual, expected = number_calls_fitness_function(fitness_batch_size=10)
    assert actual == expected

def test_number_calls_fitness_function_keep_elitism_batch_4():
    actual, expected = number_calls_fitness_function(keep_elitism=3, 
                                                     keep_parents=0,
                                                     fitness_batch_size=4)
    assert actual == expected

def test_number_calls_fitness_function_keep_parents_batch_4():
    actual, expected = number_calls_fitness_function(keep_elitism=0, 
                                                     keep_parents=4,
                                                     fitness_batch_size=4)
    assert actual == expected

def test_number_calls_fitness_function_both_keep_batch_4():
    actual, expected = number_calls_fitness_function(keep_elitism=3, 
                                                     keep_parents=4,
                                                     fitness_batch_size=4)
    assert actual == expected

def test_number_calls_fitness_function_no_keep_adaptive_mutation_batch_4():
    actual, expected = number_calls_fitness_function(keep_elitism=0, 
                                                     keep_parents=0,
                                                     mutation_type="adaptive",
                                                     mutation_percent_genes=[10, 5],
                                                     fitness_batch_size=4)
    assert actual == expected

def test_number_calls_fitness_function_default_adaptive_mutation_batch_4():
    actual, expected = number_calls_fitness_function(mutation_type="adaptive",
                                                     mutation_percent_genes=[10, 5],
                                                     fitness_batch_size=4)
    assert actual == expected

def test_number_calls_fitness_function_both_keep_adaptive_mutation_batch_4():
    actual, expected = number_calls_fitness_function(keep_elitism=3, 
                                                     keep_parents=4,
                                                     mutation_type="adaptive",
                                                     mutation_percent_genes=[10, 5],
                                                     fitness_batch_size=4)
    assert actual == expected

#### Multi Objective
def test_number_calls_fitness_function_no_keep_multi_objective_batch_1():
    actual, expected = number_calls_fitness_function(keep_elitism=0, 
                                                     keep_parents=0,
                                                     multi_objective=True,
                                                     fitness_batch_size=1)
    assert actual == expected

def test_number_calls_fitness_function_no_keep_multi_objective_batch_4():
    actual, expected = number_calls_fitness_function(keep_elitism=0, 
                                                     keep_parents=0,
                                                     multi_objective=True,
                                                     fitness_batch_size=4)
    assert actual == expected

def test_number_calls_fitness_function_no_keep_multi_objective_batch_9():
    actual, expected = number_calls_fitness_function(keep_elitism=0, 
                                                     keep_parents=0,
                                                     multi_objective=True,
                                                     fitness_batch_size=9)
    assert actual == expected

def test_number_calls_fitness_function_no_keep_multi_objective_batch_10():
    actual, expected = number_calls_fitness_function(keep_elitism=0, 
                                                     keep_parents=0,
                                                     multi_objective=True,
                                                     fitness_batch_size=10)
    assert actual == expected

def test_number_calls_fitness_function_keep_elitism_multi_objective_batch_4():
    actual, expected = number_calls_fitness_function(keep_elitism=3, 
                                                     keep_parents=0,
                                                     multi_objective=True,
                                                     fitness_batch_size=4)
    assert actual == expected

def test_number_calls_fitness_function_keep_parents_multi_objective_batch_4():
    actual, expected = number_calls_fitness_function(keep_elitism=0, 
                                                     keep_parents=4,
                                                     multi_objective=True,
                                                     fitness_batch_size=4)
    assert actual == expected

def test_number_calls_fitness_function_both_keep_multi_objective_batch_4():
    actual, expected = number_calls_fitness_function(keep_elitism=3, 
                                                     keep_parents=4,
                                                     multi_objective=True,
                                                     fitness_batch_size=4)
    assert actual == expected

def test_number_calls_fitness_function_no_keep_adaptive_mutation_multi_objective_batch_4():
    actual, expected = number_calls_fitness_function(keep_elitism=0, 
                                                     keep_parents=0,
                                                     mutation_type="adaptive",
                                                     mutation_percent_genes=[10, 5],
                                                     multi_objective=True,
                                                     fitness_batch_size=4)
    assert actual == expected

def test_number_calls_fitness_function_default_adaptive_mutation_multi_objective_batch_4():
    actual, expected = number_calls_fitness_function(mutation_type="adaptive",
                                                     mutation_percent_genes=[10, 5],
                                                     multi_objective=True,
                                                     fitness_batch_size=4)
    assert actual == expected

def test_number_calls_fitness_function_both_keep_adaptive_mutation_multi_objective_batch_4():
    actual, expected = number_calls_fitness_function(keep_elitism=3, 
                                                     keep_parents=4,
                                                     mutation_type="adaptive",
                                                     mutation_percent_genes=[10, 5],
                                                     multi_objective=True,
                                                     fitness_batch_size=4)
    assert actual == expected

#### Multi Objective NSGA-II Parent Selection
def test_number_calls_fitness_function_no_keep_multi_objective_nsga2_batch_1():
    actual, expected = number_calls_fitness_function(keep_elitism=0, 
                                                     keep_parents=0,
                                                     multi_objective=True,
                                                     fitness_batch_size=1)
    assert actual == expected

def test_number_calls_fitness_function_no_keep_multi_objective_nsga2_batch_4():
    actual, expected = number_calls_fitness_function(keep_elitism=0, 
                                                     keep_parents=0,
                                                     multi_objective=True,
                                                     fitness_batch_size=4)
    assert actual == expected

def test_number_calls_fitness_function_no_keep_multi_objective_nsga2_batch_9():
    actual, expected = number_calls_fitness_function(keep_elitism=0, 
                                                     keep_parents=0,
                                                     multi_objective=True,
                                                     fitness_batch_size=9)
    assert actual == expected

def test_number_calls_fitness_function_no_keep_multi_objective_nsga2_batch_10():
    actual, expected = number_calls_fitness_function(keep_elitism=0, 
                                                     keep_parents=0,
                                                     multi_objective=True,
                                                     fitness_batch_size=10)
    assert actual == expected

def test_number_calls_fitness_function_keep_elitism_multi_objective_nsga2_batch_4():
    actual, expected = number_calls_fitness_function(keep_elitism=3, 
                                                     keep_parents=0,
                                                     multi_objective=True,
                                                     fitness_batch_size=4)
    assert actual == expected

def test_number_calls_fitness_function_keep_parents_multi_objective_nsga2_batch_4():
    actual, expected = number_calls_fitness_function(keep_elitism=0, 
                                                     keep_parents=4,
                                                     multi_objective=True,
                                                     fitness_batch_size=4)
    assert actual == expected

def test_number_calls_fitness_function_both_keep_multi_objective_nsga2_batch_4():
    actual, expected = number_calls_fitness_function(keep_elitism=3, 
                                                     keep_parents=4,
                                                     multi_objective=True,
                                                     fitness_batch_size=4)
    assert actual == expected

def test_number_calls_fitness_function_no_keep_adaptive_mutation_multi_objective_nsga2_batch_4():
    actual, expected = number_calls_fitness_function(keep_elitism=0, 
                                                     keep_parents=0,
                                                     mutation_type="adaptive",
                                                     mutation_percent_genes=[10, 5],
                                                     multi_objective=True,
                                                     fitness_batch_size=4)
    assert actual == expected

def test_number_calls_fitness_function_default_adaptive_mutation_multi_objective_nsga2_batch_4():
    actual, expected = number_calls_fitness_function(mutation_type="adaptive",
                                                     mutation_percent_genes=[10, 5],
                                                     multi_objective=True,
                                                     fitness_batch_size=4)
    assert actual == expected

def test_number_calls_fitness_function_both_keep_adaptive_mutation_multi_objective_nsga2_batch_4():
    actual, expected = number_calls_fitness_function(keep_elitism=3, 
                                                     keep_parents=4,
                                                     mutation_type="adaptive",
                                                     mutation_percent_genes=[10, 5],
                                                     multi_objective=True,
                                                     fitness_batch_size=4)
    assert actual == expected


if __name__ == "__main__":
    #### Single-objective
    print()
    test_number_calls_fitness_function_default_keep()
    print()
    test_number_calls_fitness_function_no_keep()
    print()
    test_number_calls_fitness_function_keep_elitism()
    print()
    test_number_calls_fitness_function_keep_parents()
    print()
    test_number_calls_fitness_function_both_keep()
    print()
    test_number_calls_fitness_function_no_keep_adaptive_mutation()
    print()
    test_number_calls_fitness_function_default_adaptive_mutation()
    print()
    test_number_calls_fitness_function_both_keep_adaptive_mutation()
    print()

    #### Multi-Objective
    print()
    test_number_calls_fitness_function_no_keep_multi_objective()
    print()
    test_number_calls_fitness_function_keep_elitism_multi_objective()
    print()
    test_number_calls_fitness_function_keep_parents_multi_objective()
    print()
    test_number_calls_fitness_function_both_keep_multi_objective()
    print()
    test_number_calls_fitness_function_no_keep_adaptive_mutation_multi_objective()
    print()
    test_number_calls_fitness_function_default_adaptive_mutation_multi_objective()
    print()
    test_number_calls_fitness_function_both_keep_adaptive_mutation_multi_objective()
    print()

    #### Multi-Objective NSGA-II Parent Selection
    print()
    test_number_calls_fitness_function_no_keep_multi_objective_nsga2()
    print()
    test_number_calls_fitness_function_keep_elitism_multi_objective_nsga2()
    print()
    test_number_calls_fitness_function_keep_parents_multi_objective_nsga2()
    print()
    test_number_calls_fitness_function_both_keep_multi_objective_nsga2()
    print()
    test_number_calls_fitness_function_no_keep_adaptive_mutation_multi_objective_nsga2()
    print()
    test_number_calls_fitness_function_default_adaptive_mutation_multi_objective_nsga2()
    print()
    test_number_calls_fitness_function_both_keep_adaptive_mutation_multi_objective_nsga2()
    print()


    ######## Batch Fitness
    #### Single-objective
    print()
    test_number_calls_fitness_function_no_keep_batch_1()
    print()
    test_number_calls_fitness_function_no_keep_batch_4()
    print()
    test_number_calls_fitness_function_no_keep_batch_9()
    print()
    test_number_calls_fitness_function_no_keep_batch_10()
    print()
    test_number_calls_fitness_function_keep_elitism_batch_4()
    print()
    test_number_calls_fitness_function_keep_parents_batch_4()
    print()
    test_number_calls_fitness_function_both_keep_batch_4()
    print()
    test_number_calls_fitness_function_no_keep_adaptive_mutation_batch_4()
    print()
    test_number_calls_fitness_function_default_adaptive_mutation_batch_4()
    print()
    test_number_calls_fitness_function_both_keep_adaptive_mutation_batch_4()
    print()

    #### Multi-Objective
    print()
    test_number_calls_fitness_function_no_keep_multi_objective_batch_1()
    print()
    test_number_calls_fitness_function_no_keep_multi_objective_batch_4()
    print()
    test_number_calls_fitness_function_no_keep_multi_objective_batch_9()
    print()
    test_number_calls_fitness_function_no_keep_multi_objective_batch_10()
    print()
    test_number_calls_fitness_function_keep_elitism_multi_objective_batch_4()
    print()
    test_number_calls_fitness_function_keep_parents_multi_objective_batch_4()
    print()
    test_number_calls_fitness_function_both_keep_multi_objective_batch_4()
    print()
    test_number_calls_fitness_function_no_keep_adaptive_mutation_multi_objective_batch_4()
    print()
    test_number_calls_fitness_function_default_adaptive_mutation_multi_objective_batch_4()
    print()
    test_number_calls_fitness_function_both_keep_adaptive_mutation_multi_objective_batch_4()
    print()

    #### Multi-Objective NSGA-II Parent Selection
    print()
    test_number_calls_fitness_function_no_keep_multi_objective_nsga2_batch_1()
    print()
    test_number_calls_fitness_function_no_keep_multi_objective_nsga2_batch_4()
    print()
    test_number_calls_fitness_function_no_keep_multi_objective_nsga2_batch_9()
    print()
    test_number_calls_fitness_function_no_keep_multi_objective_nsga2_batch_10()
    print()
    test_number_calls_fitness_function_keep_elitism_multi_objective_nsga2_batch_4()
    print()
    test_number_calls_fitness_function_keep_parents_multi_objective_nsga2_batch_4()
    print()
    test_number_calls_fitness_function_both_keep_multi_objective_nsga2_batch_4()
    print()
    test_number_calls_fitness_function_no_keep_adaptive_mutation_multi_objective_nsga2_batch_4()
    print()
    test_number_calls_fitness_function_default_adaptive_mutation_multi_objective_nsga2_batch_4()
    print()
    test_number_calls_fitness_function_both_keep_adaptive_mutation_multi_objective_nsga2_batch_4()
    print()

